@@ -1,12 +1,12 @@
 # -*- coding: utf-8 -*-
 import sys
-PY3 = sys.version > '3'
 import os
 import time
 import uuid
 import codecs
 import logging
 import subprocess
+import re
 from functools import wraps
 try:
     from cookielib import Cookie, LWPCookieJar
@@ -14,78 +14,43 @@ except ImportError:
     from http.cookiejar import Cookie, LWPCookieJar
 from contextlib import contextmanager
 from .logger import configure
+from .bindings import (
+    binding,
+    QtCore,
+    QSize,
+    QByteArray,
+    QUrl,
+    QDateTime,
+    QtCriticalMsg,
+    QtDebugMsg,
+    QtFatalMsg,
+    QtWarningMsg,
+    qInstallMsgHandler,
+    QApplication,
+    QImage,
+    QPainter,
+    QPrinter,
+    QRegion,
+    QtNetwork,
+    QNetworkRequest,
+    QNetworkAccessManager,
+    QNetworkCookieJar,
+    QNetworkProxy,
+    QNetworkCookie,
+    QSslConfiguration,
+    QSsl,
+    QtWebKit,
+)
 
-__version__ = "0.1.2"
+__version__ = "0.2.3"
+
+
+PY3 = sys.version > '3'
 
 if PY3:
     unicode = str
     long = int
-
-
-bindings = ["PySide", "PyQt4"]
-binding = None
-
-
-for name in bindings:
-    try:
-        binding = __import__(name)
-        if name == 'PyQt4':
-            import sip
-            sip.setapi('QVariant', 2)
-
-    except ImportError:
-        continue
-    break
-
-
-class LazyBinding(object):
-    class __metaclass__(type):
-        def __getattr__(self, name):
-            return self.__class__
-
-    def __getattr__(self, name):
-        return self.__class__
-
-
-def _import(name):
-    if binding is None:
-        return LazyBinding()
-
-    name = "%s.%s" % (binding.__name__, name)
-    module = __import__(name)
-    for n in name.split(".")[1:]:
-        module = getattr(module, n)
-    return module
-
-
-QtCore = _import("QtCore")
-QSize = QtCore.QSize
-QByteArray = QtCore.QByteArray
-QUrl = QtCore.QUrl
-QDateTime = QtCore.QDateTime
-QtCriticalMsg = QtCore.QtCriticalMsg
-QtDebugMsg = QtCore.QtDebugMsg
-QtFatalMsg = QtCore.QtFatalMsg
-QtWarningMsg = QtCore.QtWarningMsg
-qInstallMsgHandler = QtCore.qInstallMsgHandler
-
-QtGui = _import("QtGui")
-QApplication = QtGui.QApplication
-QImage = QtGui.QImage
-QPainter = QtGui.QPainter
-QPrinter = QtGui.QPrinter
-QRegion = QtGui.QRegion
-
-QtNetwork = _import("QtNetwork")
-QNetworkRequest = QtNetwork.QNetworkRequest
-QNetworkAccessManager = QtNetwork.QNetworkAccessManager
-QNetworkCookieJar = QtNetwork.QNetworkCookieJar
-QNetworkProxy = QtNetwork.QNetworkProxy
-QNetworkCookie = QtNetwork.QNetworkCookie
-QSslConfiguration = QtNetwork.QSslConfiguration
-QSsl = QtNetwork.QSsl
-
-QtWebKit = _import('QtWebKit')
+    basestring = str
 
 
 default_user_agent = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/535.2 " +\
@@ -121,13 +86,13 @@ class GhostWebPage(QtWebKit.QWebPage):
     behaviours like alert(), confirm().
     Also intercepts client side console.log().
     """
-    def __init__(self, app, ghost):
-        self.ghost = ghost
-        super(GhostWebPage, self).__init__(app)
+    def __init__(self, app, session):
+        self.session = session
+        super(GhostWebPage, self).__init__()
 
     def chooseFile(self, frame, suggested_file=None):
-        filename = Ghost._upload_file
-        self.ghost.logger.debug('Choosing file %s' % filename)
+        filename = self.session._upload_file
+        self.session.logger.debug('Choosing file %s' % filename)
         return filename
 
     def javaScriptConsoleMessage(self, message, line, source):
@@ -138,15 +103,15 @@ class GhostWebPage(QtWebKit.QWebPage):
             source,
         )
         log_type = "warn" if "Error" in message else "info"
-        getattr(self.ghost.logger, log_type)(
+        getattr(self.session.logger, log_type)(
             "%s(%d): %s" % (source or '<unknown>', line, message),
         )
 
     def javaScriptAlert(self, frame, message):
-        """Notifies ghost for alert, then pass."""
-        self.ghost._alert = message
-        self.ghost.append_popup_message(message)
-        self.ghost.logger.info("alert('%s')" % message)
+        """Notifies session for alert, then pass."""
+        self.session._alert = message
+        self.session.append_popup_message(message)
+        self.session.logger.info("alert('%s')" % message)
 
     def _get_value(self, value):
         if callable(value):
@@ -155,34 +120,34 @@ class GhostWebPage(QtWebKit.QWebPage):
         return value
 
     def javaScriptConfirm(self, frame, message):
-        """Checks if ghost is waiting for confirm, then returns the right
+        """Checks if session is waiting for confirm, then returns the right
         value.
         """
-        if self.ghost._confirm_expected is None:
+        if self.session._confirm_expected is None:
             raise Error(
                 'You must specified a value to confirm "%s"' %
                 message,
             )
-        self.ghost.append_popup_message(message)
-        value = self.ghost._confirm_expected
-        self.ghost.logger.info("confirm('%s')" % message)
+        self.session.append_popup_message(message)
+        value = self.session._confirm_expected
+        self.session.logger.info("confirm('%s')" % message)
         return self._get_value(value)
 
     def javaScriptPrompt(self, frame, message, defaultValue, result=None):
         """Checks if ghost is waiting for prompt, then enters the right
         value.
         """
-        if self.ghost._prompt_expected is None:
+        if self.session._prompt_expected is None:
             raise Error(
                 'You must specified a value for prompt "%s"' %
                 message,
             )
-        self.ghost.append_popup_message(message)
-        value = self.ghost._prompt_expected
-        self.ghost.logger.info("prompt('%s')" % message)
+        self.session.append_popup_message(message)
+        value = self.session._prompt_expected
+        self.session.logger.info("prompt('%s')" % message)
         value = self._get_value(value)
         if value == '':
-            self.ghost.logger.warn(
+            self.session.logger.warn(
                 "'%s' prompt filled with empty string" % message,
             )
 
@@ -193,7 +158,7 @@ class GhostWebPage(QtWebKit.QWebPage):
         result.append(unicode(value))
         return True
 
-    def setUserAgent(self, user_agent):
+    def set_user_agent(self, user_agent):
         self.user_agent = user_agent
 
     def userAgentForUrl(self, url):
@@ -221,8 +186,8 @@ def can_load_page(func):
 class HttpResource(object):
     """Represents an HTTP resource.
     """
-    def __init__(self, ghost, reply, content):
-        self.ghost = ghost
+    def __init__(self, session, reply, content):
+        self.session = session
         self.url = reply.url().toString()
         self.content = content
         try:
@@ -231,7 +196,7 @@ class HttpResource(object):
             self.content = content
         self.http_status = reply.attribute(
             QNetworkRequest.HttpStatusCodeAttribute)
-        self.ghost.logger.info(
+        self.session.logger.info(
             "Resource loaded: %s %s" % (self.url, self.http_status)
         )
         self.headers = {}
@@ -242,7 +207,7 @@ class HttpResource(object):
             except UnicodeDecodeError:
                 # it will lose the header value,
                 # but at least not crash the whole process
-                self.ghost.logger.error(
+                self.session.logger.error(
                     "Invalid characters in header {0}={1}".format(
                         header,
                         reply.rawHeader(header),
@@ -260,8 +225,19 @@ def replyReadyRead(reply):
 
 class NetworkAccessManager(QNetworkAccessManager):
     """Subclass QNetworkAccessManager to always cache the reply content
+
+    :param exclude_regex: A regex use to determine wich url exclude
+        when sending a request
     """
+    def __init__(self, exclude_regex=None, *args, **kwargs):
+        self._regex = re.compile(exclude_regex) if exclude_regex else None
+        super(self.__class__, self).__init__(*args, **kwargs)
+
     def createRequest(self, operation, request, data):
+        if self._regex and self._regex.findall(str(request.url().toString())):
+            return QNetworkAccessManager.createRequest(
+                self, QNetworkAccessManager.GetOperation,
+                QNetworkRequest(QUrl()))
         reply = QNetworkAccessManager.createRequest(
             self,
             operation,
@@ -274,8 +250,85 @@ class NetworkAccessManager(QNetworkAccessManager):
 
 
 class Ghost(object):
-    """Ghost manages a QWebPage.
+    """`Ghost` manages a Qt application.
 
+    :param log_level: The optional logging level.
+    :param log_handler: The optional logging handler.
+    :param plugin_path: Array with paths to plugin directories
+        (default ['/usr/lib/mozilla/plugins'])
+    :param defaults: The defaults arguments to pass to new child sessions.
+    """
+    _app = None
+
+    def __init__(
+        self,
+        log_level=logging.WARNING,
+        log_handler=logging.StreamHandler(sys.stderr),
+        plugin_path=['/usr/lib/mozilla/plugins', ],
+        defaults=None,
+    ):
+        if not binding:
+            raise Exception("Ghost.py requires PySide or PyQt4")
+
+        self.logger = configure(
+            'ghost',
+            "Ghost",
+            log_level,
+            log_handler,
+        )
+
+        if (
+            sys.platform.startswith('linux') and
+            'DISPLAY' not in os.environ
+        ):
+            try:
+                os.environ['DISPLAY'] = ':99'
+                process = ['Xvfb', ':99', '-pixdepths', '32']
+                FNULL = open(os.devnull, 'w')
+                self.xvfb = subprocess.Popen(
+                    process,
+                    stdout=FNULL,
+                    stderr=subprocess.STDOUT,
+                )
+            except OSError:
+                raise Error('Xvfb is required to a ghost run outside ' +
+                            'an X instance')
+
+        self.logger.info('Initializing QT application')
+        Ghost._app = QApplication.instance() or QApplication(['ghost'])
+
+        qInstallMsgHandler(QTMessageProxy(
+            configure(
+                'qt',
+                'QT',
+                log_level,
+                log_handler,
+            )
+        ))
+        if plugin_path:
+            for p in plugin_path:
+                Ghost._app.addLibraryPath(p)
+
+        self.defaults = defaults or dict()
+
+    def exit(self):
+        self._app.quit()
+        if hasattr(self, 'xvfb'):
+            self.xvfb.terminate()
+
+    def start(self, **kwargs):
+        """Starts a new `Session`."""
+        kwargs.update(self.defaults)
+        return Session(self, **kwargs)
+
+    def __del__(self):
+        self.exit()
+
+
+class Session(object):
+    """`Session` manages a QWebPage.
+
+    :param ghost: The parent `Ghost` instance.
     :param user_agent: The default User-Agent header.
     :param wait_timeout: Maximum step duration in second.
     :param wait_callback: An optional callable that is periodically
@@ -287,9 +340,9 @@ class Ghost(object):
     :param ignore_ssl_errors: A boolean that forces ignore ssl errors.
     :param plugins_enabled: Enable plugins (like Flash).
     :param java_enabled: Enable Java JRE.
-    :param plugin_path: Array with paths to plugin directories
-        (default ['/usr/lib/mozilla/plugins'])
     :param download_images: Indicate if the browser should download images
+    :param exclude: A regex use to determine which url exclude
+        when sending a request
     """
     _alert = None
     _confirm_expected = None
@@ -299,79 +352,49 @@ class Ghost(object):
 
     def __init__(
         self,
+        ghost,
         user_agent=default_user_agent,
         wait_timeout=8,
         wait_callback=None,
-        log_level=logging.WARNING,
-        log_handler=logging.StreamHandler(sys.stderr),
         display=False,
         viewport_size=(800, 600),
         ignore_ssl_errors=True,
         plugins_enabled=False,
         java_enabled=False,
         javascript_enabled=True,
-        plugin_path=['/usr/lib/mozilla/plugins', ],
         download_images=True,
         show_scrollbars=True,
+        exclude=None,
         network_access_manager_class=NetworkAccessManager,
         web_page_class=GhostWebPage,
     ):
-        if not binding:
-            raise Exception("Ghost.py requires PySide or PyQt4")
+        self.ghost = ghost
 
         self.id = str(uuid.uuid4())
 
         self.logger = configure(
-            'ghost',
-            "Ghost(<%s>)" % self.id,
-            log_level,
-            log_handler,
+            'ghost.%s' % self.id,
+            "Ghost<%s>" % self.id,
+            ghost.logger.level,
         )
 
+        self.logger.info("Starting new session")
+
         self.http_resources = []
 
-        self.user_agent = user_agent
         self.wait_timeout = wait_timeout
         self.wait_callback = wait_callback
         self.ignore_ssl_errors = ignore_ssl_errors
         self.loaded = True
 
-        if (
-            sys.platform.startswith('linux')
-            and 'DISPLAY' not in os.environ
-            and not hasattr(Ghost, 'xvfb')
-        ):
-            try:
-                os.environ['DISPLAY'] = ':99'
-                process = ['Xvfb', ':99', '-pixdepths', '32']
-                FNULL = open(os.devnull, 'w')
-                Ghost.xvfb = subprocess.Popen(process, stdout=FNULL, stderr=subprocess.STDOUT)
-            except OSError:
-                raise Error('Xvfb is required to a ghost run outside ' +
-                            'an X instance')
-
         self.display = display
 
-        if not Ghost._app:
-            self.logger.info('Initializing QT application')
-            Ghost._app = QApplication.instance() or QApplication(['ghost'])
-            qInstallMsgHandler(QTMessageProxy(
-                configure(
-                    'qt',
-                    'QT',
-                    log_level,
-                    log_handler,
-                )
-            ))
-            if plugin_path:
-                for p in plugin_path:
-                    Ghost._app.addLibraryPath(p)
-
         self.popup_messages = []
-        self.page = web_page_class(Ghost._app, self)
+        self.page = web_page_class(self.ghost._app, self)
 
         if network_access_manager_class is not None:
-            self.page.setNetworkAccessManager(network_access_manager_class())
+            self.page.setNetworkAccessManager(
+                network_access_manager_class(exclude_regex=exclude))
 
         QtWebKit.QWebSettings.setMaximumPagesInCache(0)
         QtWebKit.QWebSettings.setObjectCacheCapacities(0, 0, 0)
@@ -410,11 +433,13 @@ class Ghost(object):
         self.manager = self.page.networkAccessManager()
         self.manager.finished.connect(self._request_ended)
         self.manager.sslErrors.connect(self._on_manager_ssl_errors)
+
         # Cookie jar
         self.cookie_jar = QNetworkCookieJar()
         self.manager.setCookieJar(self.cookie_jar)
+
         # User Agent
-        self.page.setUserAgent(self.user_agent)
+        self.page.set_user_agent(user_agent)
 
         self.page.networkAccessManager().authenticationRequired\
             .connect(self._authenticate)
@@ -441,9 +466,6 @@ class Ghost(object):
         if self.display:
             self.show()
 
-    def __del__(self):
-        self.exit()
-
     def frame(self, selector=None):
         """ Set main frame as current main frame's parent.
 
@@ -518,6 +540,7 @@ class Ghost(object):
                 return None
         else:
             self.page.setViewportSize(self.main_frame.contentsSize())
+
         self.logger.info("Frame size -> " + str(self.page.viewportSize()))
 
         image = QImage(self.page.viewportSize(), format)
@@ -594,7 +617,8 @@ class Ghost(object):
         printer.setOutputFormat(QPrinter.PdfFormat)
         printer.setPaperSize(QtCore.QSizeF(*paper_size), paper_units)
         printer.setPageMargins(*(paper_margins + (paper_units,)))
-        printer.setFullPage(True)
+        if paper_margins != (0, 0, 0, 0):
+            printer.setFullPage(True)
         printer.setOutputFileName(path)
         if self.webview is None:
             self.webview = QtWebKit.QWebView()
@@ -603,10 +627,14 @@ class Ghost(object):
         self.webview.print_(printer)
 
     @can_load_page
-    def click(self, selector):
+    def click(self, selector, btn=0):
         """Click the targeted element.
 
         :param selector: A CSS3 selector to targeted element.
+        :param btn: The number of mouse button.
+            0 - left button,
+            1 - middle button,
+            2 - right button
         """
         if not self.exists(selector):
             raise Error("Can't find element to click")
@@ -615,10 +643,10 @@ class Ghost(object):
                 var element = document.querySelector(%s);
                 var evt = document.createEvent("MouseEvents");
                 evt.initMouseEvent("click", true, true, window, 1, 1, 1, 1, 1,
-                    false, false, false, false, 0, element);
+                    false, false, false, false, %s, element);
                 return element.dispatchEvent(evt);
             })();
-        """ % repr(selector))
+        """ % (repr(selector), str(btn)))
 
     @contextmanager
     def confirm(self, confirm=True):
@@ -683,15 +711,14 @@ class Ghost(object):
         return not self.main_frame.findFirstElement(selector).isNull()
 
     def exit(self):
-        """Exits application and related."""
-        if self.display:
-            self.webview.close()
-        Ghost._app.quit()
+        """Exits all Qt widgets."""
+        self.logger.info("Closing session")
+        del self.webview
+        del self.cookie_jar
         del self.manager
-        del self.page
         del self.main_frame
-        if hasattr(self, 'xvfb'):
-            self.xvfb.terminate()
+        del self.page
+        self.sleep()
 
     @can_load_page
     def fill(self, selector, values):
@@ -790,6 +817,8 @@ class Ghost(object):
         wait=True,
         timeout=None,
         client_certificate=None,
+        encode_url=True,
+        user_agent=None,
     ):
         """Opens a web page.
 
@@ -809,6 +838,8 @@ class Ghost(object):
         :param timeout: An optional timeout.
         :param client_certificate An optional dict with "certificate_path" and
         "key_path" both paths corresponding to the certificate and key files
+        :param encode_url Set to true if the url have to be encoded
+        :param user_agent An option user agent string.
         :return: Page resource, and all loaded resources, unless wait
         is False, in which case it returns None.
         """
@@ -820,6 +851,9 @@ class Ghost(object):
         except AttributeError:
             raise Error("Invalid http method %s" % method)
 
+        if user_agent is not None:
+            self.page.set_user_agent(user_agent)
+
         if client_certificate:
             ssl_conf = QSslConfiguration.defaultConfiguration()
 
@@ -846,7 +880,10 @@ class Ghost(object):
 
             QSslConfiguration.setDefaultConfiguration(ssl_conf)
 
-        request = QNetworkRequest(QUrl(address))
+        if encode_url:
+            request = QNetworkRequest(QUrl(address))
+        else:
+            request = QNetworkRequest(QUrl.fromEncoded(address))
         request.CacheLoadControl(0)
         for header in headers:
             request.setRawHeader(header, headers[header])
@@ -1039,14 +1076,15 @@ class Ghost(object):
                     value,
                 )
             elif type_ == "file":
-                Ghost._upload_file = value
+                self._upload_file = value
                 res, resources = self.click(selector)
-                Ghost._upload_file = None
+
+                self._upload_file = None
         else:
-            raise Error('unsuported field tag')
+            raise Error('unsupported field tag')
 
         for event in ['input', 'change']:
-            self.fire(selector, event);
+            self.fire(selector, event)
 
         if blur:
             self.call(selector, 'blur')
@@ -1093,8 +1131,8 @@ class Ghost(object):
             self.manager.setProxy(proxy)
         else:
             raise ValueError(
-                'Unsupported proxy type:' + type_
-                + '\nsupported types are: none/socks5/http/https/default',
+                'Unsupported proxy type: %s' % type_ +
+                '\nsupported types are: none/socks5/http/https/default',
             )
 
     def set_viewport_size(self, width, height):
@@ -1120,7 +1158,7 @@ class Ghost(object):
 
         while time.time() <= (started_at + value):
             time.sleep(0.01)
-            Ghost._app.processEvents()
+            self.ghost._app.processEvents()
 
     def wait_for(self, condition, timeout_message, timeout=None):
         """Waits until condition is True.
